//This is the source code for the command line executable supplied
//with the tutorial "Secrets of simple image filtering" by Christopher
//Atkinson. The tutorial is available for reading at www.gamedev.net.

//You can use this source code freely for ALL puposes - there are
//no licenses or anything attached to this - unfortunately this
//includes anykind of guarantee that it will work on all (Windows)
//workstations. I also wash my hands clean of any damage this code
//might cause anybody or anything. If you DO use this code somewhere
//or find it helpful, mention me and let me know - I'd like to know
//where it ends up :). If you find a bug or have a neato
//suggestion how to improve things, also let me know at
//christopheratkinson@hotmail.co.uk.

//This code was written using the Borland BC5.02 IDE and compiler.
//The aim was to strip the amount of text down to the absolute
//minimum while maximizing simplicity. For that reason most of the
//stuff related to loading and writing a bitmap have been rewritten
//without the use of WinAPI. If you're only interested in the filtering
//part, skim down to the middle of the file where Write24BitBitmap()
//ends.

#include <fstream>
#include <mem>
#include <stdio>

//standard macros copied from the docs
#define min(a, b)  (((a) < (b)) ? (a) : (b))
#define max(a, b)  (((a) > (b)) ? (a) : (b))

//some necessary typedefs
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef long LONG;
typedef unsigned char BYTE;

//optionally redefine the biCompression values
//for cross-platform usability
#ifndef BI_RGB
	#define BI_RGB      		0L
#endif
#ifndef BI_RLE8
	#define BI_RLE8     		1L
#endif
#ifndef BI_RLE4
	#define BI_RLE4     		2L
#endif

//an alias for tagBITMAPFILEHEADER
typedef struct myBMP_FILE_HEADER {
	WORD    bfType;
   DWORD   bfSize;
   WORD    bfReserved1;
   WORD    bfReserved2;
   DWORD   bfOffBits;
	} BMP_FILE_HEADER;

//an alias for tagBITMAPINFOHEADER
typedef struct myBMP_INFO_HEADER {
   DWORD  biSize;
   LONG   biWidth;
   LONG   biHeight;
   WORD   biPlanes;
   WORD   biBitCount;
   DWORD  biCompression;
   DWORD  biSizeImage;
   LONG   biXPelsPerMeter;
   LONG   biYPelsPerMeter;
   DWORD  biClrUsed;
   DWORD  biClrImportant;
	} BMP_INFO_HEADER;

//a pixel/sample in a 24bit bitmap consists of three components
typedef struct TPixel {
	BYTE r, g, b;
	} TPixel;

//the bitmap loading routine
BYTE* Load24BitBitmap(const char * pFileName, int * pWidth, int * pHeight)
{
	BMP_FILE_HEADER BFH;
   BMP_INFO_HEADER BIH;

   FILE *f = fopen(pFileName, "rb");

   if(f == NULL)
   	{
      fclose(f);
      cout << "Cannot open " << pFileName << endl;
   	return NULL;
      }

   //read bitmap file header
	fread(&BFH.bfType, sizeof(WORD), 1, f);
		if(BFH.bfType != (('M' << 8) +  'B'))
         return NULL;
	fread(&BFH.bfSize, sizeof(DWORD), 1, f);
	fread(&BFH.bfReserved1, sizeof(WORD), 2, f);
	fread(&BFH.bfOffBits, sizeof(DWORD), 1, f);

	//read bitmap info header
	fread(&BIH.biSize, sizeof(DWORD), 1, f);
	fread(&BIH.biWidth, sizeof(LONG), 1, f);
	fread(&BIH.biHeight, sizeof(LONG), 1, f);
	fread(&BIH.biPlanes, sizeof(WORD), 1, f);
	fread(&BIH.biBitCount, sizeof(WORD), 1, f);
	fread(&BIH.biCompression, sizeof(DWORD), 1, f);
	fread(&BIH.biSizeImage, sizeof(DWORD), 1, f);
	fread(&BIH.biXPelsPerMeter, sizeof(LONG), 1, f);
	fread(&BIH.biYPelsPerMeter, sizeof(LONG), 1, f);
	fread(&BIH.biClrUsed, sizeof(DWORD), 1, f);
	fread(&BIH.biClrImportant, sizeof(DWORD), 1, f);

   int numChannels = 3;

   //initialize the data storage
   BYTE* data = new BYTE[BIH.biWidth * BIH.biHeight * numChannels];

   //return the width and height of the bitmap
   *pWidth = BIH.biWidth;
   *pHeight = BIH.biHeight;

   TPixel pixel;
 	int i = 0;

   //act according to the compression type used
   switch(BIH.biCompression)
	  	{
      case BI_RLE4:
      	cout << "Cannot load RLE4 encoded images" << endl;
         delete[] data;
         fclose(f);
      	return NULL;
      case BI_RLE8:
      	cout << "Cannot load RLE8 encoded images" << endl;
         delete[] data;
         fclose(f);
      	return NULL;
      //assume no compression
      case BI_RGB:
      default:
			for(int x = 0; x < BIH.biWidth; x++)
  			  	{
				for(int y = 0; y < BIH.biHeight; y++)
      	 		{
           	   //read a pixel and store it in the output array; notice how we're
               //reversing the component order from bgr to rgb - this is beacuse
               //bitmaps store the color info in bgr format  
					fread(&pixel, sizeof(TPixel), 1, f);
	         	data[i + 2] = (BYTE)pixel.r;
					data[i + 1] = (BYTE)pixel.g;
  			      data[i] = (BYTE)pixel.b;
					i += 3;
  		   		}
				}
         break;
      }

   fclose(f);

   return data;
}

//the bitmap output routine
bool Write24BitBitmap(const char * pFileName, BYTE * pData, int pWidth, int pHeight)
{
	BMP_FILE_HEADER BFH;
   BMP_INFO_HEADER BIH;

   FILE *f = fopen(pFileName, "wb");

   //set the file header info
   BFH.bfType = 'M';
   BFH.bfType <<= 8;
   BFH.bfType |= 'B';

   BFH.bfSize = (pWidth * pHeight * 3);
	BFH.bfReserved1 = 0;
	BFH.bfReserved2 = 0;
   BFH.bfOffBits = sizeof(BMP_INFO_HEADER) + sizeof(BMP_FILE_HEADER);

   //set other header information
   BIH.biSize = sizeof(BMP_INFO_HEADER);
   BIH.biWidth = pWidth;
   BIH.biHeight = pHeight;
   BIH.biPlanes = 1;
   BIH.biBitCount = 24;
   BIH.biCompression = BI_RGB;
   BIH.biSizeImage = 0;
   BIH.biXPelsPerMeter = 0;
   BIH.biYPelsPerMeter = 0;
   BIH.biClrUsed = 0;
   BIH.biClrImportant = 0;

   //write the whole thing
   fwrite(&BFH, sizeof(BMP_FILE_HEADER), 1, f);
   fwrite(&BIH, sizeof(BMP_INFO_HEADER), 1, f);
   fwrite(pData, pWidth * pHeight * 3, 1, f);

   //close the output file and return a smiley :)
   fclose(f);

   return true;
}


//				***	THE FILTER	 ***				//

//the three following declarations make up the filter: we're using a
//constant to define the negative values as typing out the kernel
//directly could easily become difficult to read; we're presuming
//the kernel is square-shaped, eg it has equal length edges defined
//by KernelSize and the kernel (or PSF) itself is a simple array
//of floating point values
const float c = -1/8.f;

int KernelSize = 3;

float Kernel[] =	{
				c, c, c,
            c, 1, c,
            c, c, c,
            };

//				***	USING THE FILTER	 ***				//

BYTE* DetectEdges(BYTE * pSourceData, int pWidth, int pHeight)
{

//STEP 1: initialize the variables

	int i, j;

   //we're doing only greyscale filtering - therefore we will be needing
   //a simple 8bit image to build off of
   static BYTE* Source8Bit = NULL;

   //the output samples fill first be in floating point format, stored in
   //Temp[]
   float* Temp = new float[pWidth * pHeight];
   Source8Bit = new BYTE[pWidth * pHeight];

   //zero out the temp array
   memset(Temp, 0, pWidth * pHeight * sizeof(float));

//STEP 2: resample the source data from 24 bits to 8 bits by discrading
//			 two components

   //make the bitmap 8-bit
   for(i = 0, j = 0; i < pWidth * pHeight * 3; i += 3, j++)
   	Source8Bit[j] = pSourceData[i];

   i = 0;
   int Value;

//STEP 3: convolving the source image with the filter kernel

   //THIS IS THE HEART OF THE WHOLE THING

   //for every column in the input image
   for(int Imgy = 0; Imgy < pHeight; Imgy++)
   	{
   	//and every row in the input image
		for(int Imgx = 0; Imgx < pWidth; Imgx++)
      	{
         //apply every column in the filter kernel
			for(int Fy = 0, y = -1; Fy < KernelSize; Fy++, y++)
         	{
            //and every row in the filter kernel
   			for(int Fx = 0, x = -1; Fx < KernelSize; Fx++, x++)
            	{
               //if the coordinates are outside of the source image,
               //give the sample a value 0
               if(Imgx + x < 0 ||
	               Imgx + x > pWidth ||
                  Imgy + y > pHeight ||
                  Imgy + y < 0)
	               Value = 0;
               //otherwise fetch the value from the input image
               else
                  Value = Source8Bit[(Imgy + y) * pWidth + (Imgx + x)];

               //multiply it with the corresponding value in the filter
               //kernel and accumulate the result
		         Temp[i] += Kernel[Fy * KernelSize + Fx] * Value;
               }
            }
         //to normalise the output, multiply the value by a constant found
         //through experimentation
			Temp[i] *= 9;
         //and clip it
         Temp[i] = min(Temp[i], 255.f);
  	      i++;
         }
      }

	//look closely at the two inmost loops, more precisely the starting values
   //of x and y. Since the PSF has to be centered over a specific sample in the
   //input image, we need to start applying the filter accordingly - if you don't
   //get it, think about a little and it'll start to make sense! For KernelSize =
   //5, this value would have to be -2
                                      
//STEP 4: returning 24-bit data

   //this is where we will store the output image - the Width and
   //Height of the output image will be the same as those of the
   //input image
   BYTE* Destination = new BYTE[pWidth * pHeight * 3];

   BYTE ColorElement;

   //copy the whole mess
	for(j = 0; j < pWidth * pHeight; j++)
		{
      //do negative sample clipping
	   ColorElement = max(Temp[j], 0.f);
  		Destination[j * 3] = Destination[j * 3 + 1] = Destination[j * 3 + 2] = ColorElement;
   	}

   return Destination;
}

void main()
{
	int width, height;

   //load the image
	BYTE* data = Load24BitBitmap("test.bmp", &width, &height);

   //do some basic error checking
   if(data == NULL)
   	{
   	cout << "An error ocurred while loading the image" << endl;
      return;
      }

   //apply the filter
   BYTE* data2 = DetectEdges(data, width, height);

   //dump the output image
   Write24BitBitmap("testout.bmp", data2, width, height);

	//clean up
   delete[] data;
   delete[] data2;
}
